#    Copyright 2022-2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Tiler for reading tiles from 3Dhistech tiff files."""

from functools import lru_cache
from pathlib import Path
from typing import Optional, Union

from tifffile.tifffile import TiffFile, TiffPageSeries
from upath import UPath

from opentile.formats.histech.histech_tiff_image import HistechTiffImage
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata
from opentile.tiff_image import TiffImage
from opentile.tiler import Tiler


class HistechTiffTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path, UPath],
        turbo_path: Optional[Union[str, Path]] = None,
    ):
        """Tiler for 3DHistech tiff file.

        Parameters
        ----------
        filepath: Union[str, Path, UPath]
            Filepath to a 3DHistech-TiffFile.
        """
        super().__init__(UPath(filepath))
        self._jpeg = Jpeg(turbo_path)

    @property
    def metadata(self) -> Metadata:
        """No known metadata for 3DHistech tiff files."""
        return Metadata()

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return "3dh_PixelSizeX" in tiff_file.pages.first.description

    @lru_cache(None)
    def get_level(self, level: int, page: int = 0) -> TiffImage:
        return HistechTiffImage(
            self._get_tiff_page(self._level_series_index, level, page),
            self._fh,
            self.base_size,
        )

    def get_label(self, page: int = 0) -> TiffImage:
        raise NotImplementedError()

    def get_overview(self, page: int = 0) -> TiffImage:
        raise NotImplementedError()

    @staticmethod
    def _is_level_series(series: TiffPageSeries) -> bool:
        return series.index == 0

    @staticmethod
    def _is_overview_series(series: TiffPageSeries) -> bool:
        """Return true if series is a overview series."""
        return False

    @staticmethod
    def _is_label_series(series: TiffPageSeries) -> bool:
        """Return true if series is a label series."""
        return False



#    Copyright 2021-2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Tiler for reading tiles from ndpi files."""

import math
from functools import lru_cache
from pathlib import Path
from typing import Optional, Union

from tifffile.tifffile import TiffFile, TiffPage, TiffPageSeries
from upath import UPath

from opentile.formats.ndpi.ndpi_image import (
    NdpiCroppedImage,
    NdpiOneFrameImage,
    NdpiStripedImage,
)
from opentile.formats.ndpi.ndpi_metadata import NdpiMetadata
from opentile.geometry import Size
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata
from opentile.tiff_image import TiffImage
from opentile.tiler import Tiler


class NdpiTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path, UPath],
        tile_size: int,
        turbo_path: Optional[Union[str, Path]] = None,
        label_crop_position: float = 0.3,
    ):
        """Tiler for ndpi file, with functions to produce tiles of specified
        size.

        Parameters
        ----------
        filepath: Union[str, Path, UPath]
            Filepath to a ndpi TiffFile.
        tile_size: int
            Tile size to cache and produce. Must be multiple of 8 and will be
            adjusted to be an even multiplier or divider of the smallest strip
            width in the file.
        turbo_path: Optional[Union[str, Path]] = None
            Path to turbojpeg (dll or so).
        label_crop_position: float = 0.3
            The position (relative to the image width) to use for cropping out
            the label and overview image from the macro image.

        """
        super().__init__(filepath)
        self._tile_size = Size(tile_size, tile_size)
        self._tile_size = self._adjust_tile_size(
            tile_size, self._get_smallest_stripe_width()
        )
        if self.tile_size.width % 8 != 0 or self.tile_size.height % 8 != 0:
            raise ValueError(f"Tile size {self.tile_size} not divisible by 8")
        self._jpeg = Jpeg(turbo_path)
        self._metadata = NdpiMetadata(self.base_page)
        self._label_crop_position = label_crop_position

    @property
    def tile_size(self) -> Size:
        """The size of the tiles to generate."""
        return self._tile_size

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return tiff_file.is_ndpi

    @staticmethod
    def _is_level_series(series: TiffPageSeries) -> bool:
        return series.index == 0

    @staticmethod
    def _is_overview_series(series: TiffPageSeries) -> bool:
        """Return true if series is a overview series."""
        return series.name == "Macro"

    @staticmethod
    def _is_label_series(series: TiffPageSeries) -> bool:
        """Return true if series is a label series."""
        return False

    @staticmethod
    def _adjust_tile_size(
        requested_tile_width: int, smallest_stripe_width: Optional[int] = None
    ) -> Size:
        """Return adjusted tile size. If file contains striped images the
        tile size must be an n * smallest stripe width in the file, where n
        is the closest square factor of the ratio between requested tile width
        and smallest stripe width.

        Parameters
        ----------
        requested_tile_width: int
            Requested tile width.
        smallest_stripe_width: Optional[int] = None
            Smallest stripe width in file.

        Returns
        ----------
        Size
            Adjusted tile size.
        """
        if (
            smallest_stripe_width is None
            or smallest_stripe_width == requested_tile_width
        ):
            # No striped pages or requested is equal to smallest
            return Size(requested_tile_width, requested_tile_width)

        if requested_tile_width > smallest_stripe_width:
            factor = requested_tile_width / smallest_stripe_width
        else:
            factor = smallest_stripe_width / requested_tile_width
        # Factor should be a square number (in the series 2^n)
        factor_2 = pow(2, round(math.log2(factor)))
        adjusted_width = factor_2 * smallest_stripe_width
        return Size(adjusted_width, adjusted_width)

    def _get_smallest_stripe_width(self) -> Optional[int]:
        """Return smallest stripe width in file, or None if no image in the
        file is striped.

        Returns
        ----------
        Optional[int]
            The smallest stripe width in the file, or None if no image in the
            file is striped.
        """
        smallest_stripe_width: Optional[int] = None
        for page in self._tiff_file.pages:
            assert isinstance(page, TiffPage)
            stripe_width = page.chunks[1]
            if page.is_tiled and (
                smallest_stripe_width is None or smallest_stripe_width > stripe_width
            ):
                smallest_stripe_width = stripe_width
        return smallest_stripe_width

    @lru_cache(None)
    def get_level(
        self,
        level: int,
        page: int = 0,
    ) -> TiffImage:
        """Create a level image.

        Parameters
        ----------
        level: int
            Level of page.
        page: int
            Page to use.

        Returns
        ----------
        TiffImage
            Created image.
        """
        tiff_page = (
            self._tiff_file.series[self._level_series_index].levels[level].pages[page]
        )
        assert isinstance(tiff_page, TiffPage)
        if tiff_page.is_tiled:  # Striped ndpi page
            return NdpiStripedImage(
                tiff_page, self._fh, self.base_size, self.tile_size, self._jpeg
            )
        # Single frame, force tiling
        return NdpiOneFrameImage(
            tiff_page, self._fh, self.base_size, self.tile_size, self._jpeg
        )

    @lru_cache(None)
    def get_label(self, page: int = 0) -> TiffImage:
        """Create a label image.

        Returns
        ----------
        TiffImage
            Created image.
        """
        assert self._overview_series_index is not None
        tiff_page = self._tiff_file.series[self._overview_series_index].pages.pages[
            page
        ]
        assert isinstance(tiff_page, TiffPage)
        return NdpiCroppedImage(
            tiff_page, self._fh, self._jpeg, (0.0, self._label_crop_position)
        )

    @lru_cache(None)
    def get_overview(self, page: int = 0) -> TiffImage:
        """Create a overview image.

        Returns
        ----------
        TiffImage
            Created image.
        """
        assert self._overview_series_index is not None
        tiff_page = self._tiff_file.series[self._overview_series_index].pages.pages[
            page
        ]
        assert isinstance(tiff_page, TiffPage)
        return NdpiCroppedImage(
            tiff_page, self._fh, self._jpeg, (self._label_crop_position, 1.0)
        )



#    Copyright 2021-2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Geometry classes."""

import math
from dataclasses import dataclass
from typing import Generator, Sequence, Tuple, Union


@dataclass
class SizeMm:
    width: float
    height: float

    def __str__(self) -> str:
        return f"{self.width}x{self.height}"

    def __mul__(self, factor: Union[int, float, "Size", "Point"]) -> "SizeMm":
        if isinstance(factor, (int, float)):
            return SizeMm(factor * self.width, factor * self.height)
        elif isinstance(factor, Size):
            return SizeMm(factor.width * self.width, factor.height * self.height)
        elif isinstance(factor, Point):
            return SizeMm(factor.x * self.width, factor.y * self.height)
        return NotImplemented

    def __truediv__(self, divider: Union[int, float, "SizeMm"]) -> "SizeMm":
        """Return divided SizeMm without rounding."""
        if isinstance(divider, (int, float)):
            return SizeMm(self.width / divider, self.height / divider)
        elif isinstance(divider, SizeMm):
            return SizeMm(self.width / divider.width, self.height / divider.height)
        return NotImplemented

    def __floordiv__(self, divider: Union[int, float, "SizeMm"]) -> "SizeMm":
        """Return divided SizeMm rounded down to closest integer width and
        height.
        """
        if isinstance(divider, (int, float)):
            return SizeMm(int(self.width / divider), int(self.height / divider))
        if isinstance(divider, SizeMm):
            return SizeMm(
                int(self.width / divider.width), int(self.height / divider.height)
            )
        return NotImplemented

    def to_int_tuple(self) -> Tuple[int, int]:
        return int(self.width), int(self.height)

    def to_tuple(self) -> Tuple[float, float]:
        return self.width, self.height

    @classmethod
    def from_tuple(cls, input: Union[Tuple[float, float], Sequence[float]]) -> "SizeMm":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")


@dataclass
class PointMm:
    x: float
    y: float

    def __str__(self) -> str:
        return f"{self.x},{self.y}"

    def __floordiv__(self, divider: SizeMm) -> "Point":
        """Return divided PointMm rounded down to closest integer x and y."""
        if isinstance(divider, SizeMm):
            return Point(int(self.x / divider.width), int(self.y / divider.height))
        return NotImplemented

    def __add__(self, value: Union[int, float, SizeMm, "PointMm"]) -> "PointMm":
        if isinstance(value, (int, float)):
            return PointMm(self.x + value, self.y + value)
        elif isinstance(value, SizeMm):
            return PointMm(self.x + value.width, self.y + value.height)
        elif isinstance(value, PointMm):
            return PointMm(self.x + value.x, self.y + value.y)
        return NotImplemented

    @classmethod
    def from_tuple(
        cls, input: Union[Tuple[float, float], Sequence[float]]
    ) -> "PointMm":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")


@dataclass
class Size:
    width: int
    height: int

    def __str__(self) -> str:
        return f"{self.width}x{self.height}"

    def __neg__(self) -> "Size":
        return Size(-self.width, -self.height)

    def __sub__(self, value: Union[int, "Size", "Point"]) -> "Size":
        if isinstance(value, int):
            return Size(self.width - value, self.height - value)
        elif isinstance(value, Size):
            return Size(self.width - value.width, self.height - value.height)
        elif isinstance(value, Point):
            return Size(self.width - value.x, self.height - value.y)
        return NotImplemented

    def __add__(self, value: Union[int, "Size", "Point"]) -> "Size":
        if isinstance(value, int):
            return Size(self.width + value, self.height + value)
        elif isinstance(value, Size):
            return Size(self.width + value.width, self.height + value.height)
        elif isinstance(value, Point):
            return Size(self.width + value.x, self.height + value.y)
        return NotImplemented

    def __mul__(self, factor: Union[int, float, "Size", "Point"]) -> "Size":
        if isinstance(factor, (int, float)):
            return Size(int(factor * self.width), int(factor * self.height))
        elif isinstance(factor, Size):
            return Size(factor.width * self.width, factor.height * self.height)
        elif isinstance(factor, Point):
            return Size(factor.x * self.width, factor.y * self.height)
        return NotImplemented

    def __floordiv__(self, divider: Union[int, float, "Size", SizeMm]) -> "Size":
        """Return divided Size rounded down to closest integer width and
        height.
        """
        if isinstance(divider, (int, float)):
            return Size(int(self.width / divider), int(self.height / divider))
        elif isinstance(divider, (Size, SizeMm)):
            return Size(
                int(self.width / divider.width), int(self.height / divider.height)
            )
        return NotImplemented

    def ceil_div(self, divider: Union[int, float, "Size", SizeMm]) -> "Size":
        """Return divided Size rounded up to closest integer width and height."""
        if isinstance(divider, (int, float)):
            return Size(
                math.ceil(self.width / divider), math.ceil(self.height / divider)
            )
        elif isinstance(divider, (Size, SizeMm)):
            return Size(
                math.ceil(self.width / divider.width),
                math.ceil(self.height / divider.height),
            )
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.width, self.height))

    def __lt__(self, item: "Size") -> bool:
        if isinstance(item, Size):
            return self.width < item.width
        return NotImplemented

    @classmethod
    def from_points(cls, point_1: "Point", point_2: "Point") -> "Size":
        return cls(point_2.x - point_1.x, point_2.y - point_1.y)

    def to_tuple(self) -> Tuple[int, int]:
        return (self.width, self.height)

    @classmethod
    def from_tuple(cls, input: Union[Tuple[int, int], Sequence[int]]) -> "Size":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")

    @classmethod
    def max(cls, size_1: "Size", size_2: "Size") -> "Size":
        return cls(
            width=max(size_1.width, size_2.width),
            height=max(size_1.height, size_2.height),
        )

    def ceil(self) -> "Size":
        return Size(
            width=int(math.ceil(self.width)), height=int(math.ceil(self.height))
        )

    @property
    def area(self) -> int:
        return self.width * self.height


@dataclass
class Point:
    x: int
    y: int

    def __str__(self) -> str:
        return f"{self.x},{self.y}"

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def __mul__(self, factor: Union[int, float, Size, "Point"]) -> "Point":
        if isinstance(factor, (int, float)):
            return Point(int(factor * self.x), int(factor * self.y))
        elif isinstance(factor, Size):
            return Point(factor.width * self.x, factor.height * self.y)
        elif isinstance(factor, Point):
            return Point(factor.x * self.x, factor.y * self.y)
        return NotImplemented

    def __floordiv__(
        self, divider: Union[int, float, Size, SizeMm, "Point"]
    ) -> "Point":
        """Return divided Point rounded down to closest integer x and y."""
        if isinstance(divider, (int, float)):
            return Point(int(self.x / divider), int(self.y / divider))
        elif isinstance(divider, Point):
            return Point(int(self.x / divider.x), int(self.y / divider.y))
        elif isinstance(divider, (Size, SizeMm)):
            return Point(int(self.x / divider.width), int(self.y / divider.height))
        return NotImplemented

    def ceil_div(self, divider: Union[int, float, "Point", Size, SizeMm]) -> "Point":
        """Return divided Point rounded up closest integer x and y."""
        if isinstance(divider, (int, float)):
            return Point(math.ceil(self.x / divider), math.ceil(self.y / divider))
        elif isinstance(divider, Point):
            return Point(math.ceil(self.x / divider.x), math.ceil(self.y / divider.y))
        elif isinstance(divider, (Size, SizeMm)):
            return Point(
                math.ceil(self.x / divider.width), math.ceil(self.y / divider.height)
            )
        return NotImplemented

    def __mod__(self, divider: Union[Size, "Point"]) -> "Point":
        if isinstance(divider, Size):
            return Point(self.x % divider.width, self.y % divider.height)
        elif isinstance(divider, Point):
            return Point(self.x % divider.x, self.y % divider.y)
        return NotImplemented

    def __add__(self, value: Union[int, float, Size, "Point"]) -> "Point":
        if isinstance(value, (int, float)):
            return Point(int(self.x + value), int(self.y + value))
        elif isinstance(value, Size):
            return Point(self.x + value.width, self.y + value.height)
        elif isinstance(value, Point):
            return Point(self.x + value.x, self.y + value.y)
        return NotImplemented

    def __sub__(self, value: Union[int, float, Size, "Point"]) -> "Point":
        if isinstance(value, (int, float)):
            return Point(int(self.x - value), int(self.y - value))
        elif isinstance(value, Size):
            return Point(self.x - value.width, self.y - value.height)
        elif isinstance(value, Point):
            return Point(self.x - value.x, self.y - value.y)
        return NotImplemented

    @classmethod
    def max(cls, point_1: "Point", point_2: "Point") -> "Point":
        return cls(x=max(point_1.x, point_2.x), y=max(point_1.y, point_2.y))

    @classmethod
    def min(cls, point_1: "Point", point_2: "Point") -> "Point":
        return cls(x=min(point_1.x, point_2.x), y=min(point_1.y, point_2.y))

    def to_tuple(self) -> Tuple[int, int]:
        return (self.x, self.y)

    @classmethod
    def from_tuple(cls, input: Union[Tuple[int, int], Sequence[int]]) -> "Point":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")


@dataclass
class Region:
    position: Point
    size: Size

    def __str__(self) -> str:
        return f"from {self.start} to {self.end}"

    @property
    def start(self) -> Point:
        return self.position

    @property
    def end(self) -> Point:
        end: Point = self.position + self.size
        return end

    def __mul__(self, value: int) -> "Region":
        if isinstance(value, int):
            return Region(position=self.position * value, size=self.size * value)
        return NotImplemented

    @property
    def box(self) -> Tuple[int, int, int, int]:
        return self.start.x, self.start.y, self.end.x, self.end.y

    @property
    def box_from_origin(self) -> Tuple[int, int, int, int]:
        return 0, 0, self.size.width, self.size.height

    def iterate_all(
        self,
    ) -> Generator[Point, None, None]:
        return (
            Point(x, y)
            for y in range(self.start.y, self.end.y)
            for x in range(self.start.x, self.end.x)
        )

    @classmethod
    def from_points(cls, point_1: "Point", point_2: "Point") -> "Region":
        return cls(
            position=point_1,
            size=Size(width=point_2.x - point_1.x, height=point_2.y - point_1.y),
        )

    @classmethod
    def from_tile(cls, tile: "Point", size: "Size"):
        return cls(position=tile * size, size=size)

    def is_inside(self, test_region: "Region") -> bool:
        return bool(
            (self.start.x >= test_region.start.x)
            and (self.start.y >= test_region.start.y)
            and (self.end.x <= test_region.end.x)
            and (self.end.y <= test_region.end.y)
        )

    def crop(self, region: "Region") -> "Region":
        start = Point.min(Point.max(region.position, self.position), self.end)
        end = Point.max(Point.min(region.end, self.end), self.position)
        size = Size.from_points(start, end)
        return Region(position=start, size=size)


@dataclass
class RegionMm:
    position: PointMm
    size: SizeMm

    def __str__(self) -> str:
        return f"from {self.start} to {self.end}"

    @property
    def start(self) -> PointMm:
        return self.position

    @property
    def end(self) -> PointMm:
        end: PointMm = self.position + self.size
        return end



#    Copyright 2021-2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Tiler for reading tiles from Philips tiff files."""

from functools import lru_cache
from pathlib import Path
from typing import Optional, Union

from tifffile.tifffile import TiffFile, TiffPage, TiffPageSeries
from upath import UPath

from opentile.formats.philips.philips_tiff_image import PhilipsTiffImage
from opentile.formats.philips.philips_tiff_metadata import PhilipsTiffMetadata
from opentile.geometry import SizeMm
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata
from opentile.tiff_image import TiffImage
from opentile.tiler import Tiler


class PhilipsTiffTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path, UPath],
        turbo_path: Optional[Union[str, Path]] = None,
    ):
        """Tiler for Philips tiff file.

        Parameters
        ----------
        filepath: Union[str, Path, UPath]
            Filepath to a Philips-TiffFile.
        turbo_path: Optional[Union[str, Path]] = None
            Path to turbojpeg (dll or so).
        """
        super().__init__(filepath)
        self._jpeg = Jpeg(turbo_path)
        self._metadata = PhilipsTiffMetadata(self._tiff_file)
        assert self._metadata.pixel_spacing is not None
        self._base_mpp = SizeMm.from_tuple(self._metadata.pixel_spacing) * 1000.0

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return tiff_file.is_philips

    def get_level(self, level: int, page: int = 0) -> TiffImage:
        return self._get_image(self._level_series_index, level, page)

    def get_label(self, page: int = 0) -> TiffImage:
        return self._get_image(self._label_series_index, 0, page)

    def get_overview(self, page: int = 0) -> TiffImage:
        return self._get_image(self._overview_series_index, 0, page)

    @lru_cache(None)
    def _get_image(self, series: int, level: int, page: int = 0) -> TiffImage:
        """Return PhilipsTiffTiledPage for series, level, page."""
        return PhilipsTiffImage(
            self._get_tiff_page(series, level, page),
            self._fh,
            self.base_size,
            self._base_mpp,
            self._jpeg,
        )

    @staticmethod
    def _is_level_series(series: TiffPageSeries) -> bool:
        return series.index == 0

    @staticmethod
    def _is_overview_series(series: TiffPageSeries) -> bool:
        """Return true if series is a overview series."""
        page = series.pages[0]
        assert isinstance(page, TiffPage)
        return page.description.find("Macro") > -1

    @staticmethod
    def _is_label_series(series: TiffPageSeries) -> bool:
        """Return true if series is a label series."""
        page = series.pages[0]
        assert isinstance(page, TiffPage)
        return page.description.find("Label") > -1

    @staticmethod
    def _get_associated_mpp_from_page(page: TiffPage):
        """Return mpp (um/pixel) for associated image (label or
        macro) from page."""
        pixel_size_start_string = "pixelsize=("
        pixel_size_start = page.description.find(pixel_size_start_string)
        pixel_size_end = page.description.find(")", pixel_size_start)
        pixel_size_string = page.description[
            pixel_size_start + len(pixel_size_start_string) : pixel_size_end
        ]
        pixel_spacing = SizeMm.from_tuple(
            [float(v) for v in pixel_size_string.replace('"', "").split(",")]
        )
        return pixel_spacing / 1000.0



#    Copyright 2021-2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Main interface for OpenTile."""

from pathlib import Path
from typing import Any, BinaryIO, Dict, Optional, Tuple, Type, Union

from tifffile import TiffFile, TiffFileError

from opentile.formats import (
    HistechTiffTiler,
    NdpiTiler,
    OmeTiffTiler,
    PhilipsTiffTiler,
    SvsTiler,
)
from fsspec.core import open
from opentile.tiler import Tiler
from upath import UPath


class OpenTile:
    @classmethod
    def open(
        cls,
        filepath: Union[str, Path, UPath],
        tile_size: int = 512,
        file_options: Optional[Dict[str, str]] = None,
        turbo_path: Optional[Union[str, Path]] = None,
    ) -> Tiler:
        """Return a file type specific tiler for tiff file in filepath.
        Tile size and turbo jpeg path are optional but required for some file
        types.

        Parameters
        ----------
        filepath: Union[str, Path, UPath]
            Path to tiff file.
        tile_size: int = 512
            Tile size for creating tiles, if needed for file format.
        file_options: Optional[Dict[str, str]] = None
            Options to pass to filesystem when opening file.
        turbo_path: Optional[Union[str, Path]] = None
            Path to turbojpeg (dll or so).
        """
        format, supported_tiler = cls.get_tiler(filepath, file_options)
        if supported_tiler is NdpiTiler:
            return NdpiTiler(filepath, tile_size, turbo_path)

        if supported_tiler is SvsTiler:
            return SvsTiler(filepath, turbo_path)

        if supported_tiler is PhilipsTiffTiler:
            return PhilipsTiffTiler(filepath, turbo_path)

        if supported_tiler is HistechTiffTiler:
            return HistechTiffTiler(filepath)

        if supported_tiler is OmeTiffTiler:
            return OmeTiffTiler(filepath)

        raise NotImplementedError("Non supported tiff file")

    @staticmethod
    def get_tiler(
        filepath: Union[str, Path, UPath],
        file_options: Optional[Dict[str, Any]] = None,
    ) -> Tuple[Optional[str], Optional[Type[Tiler]]]:
        """Return tiler that supports the tiff file in filepath, or None if
        not supported"""
        tilers: Dict[str, Type[Tiler]] = {
            "ndpi": NdpiTiler,
            "svs": SvsTiler,
            "phillips tiff": PhilipsTiffTiler,
            "3dhistech tiff": HistechTiffTiler,
            "ome-tiff tiler": OmeTiffTiler,
        }
        try:
            file: BinaryIO = open(str(filepath), **file_options or {})  # type: ignore
            tiff_file = TiffFile(file)
            return next(
                (tiler_name, tiler)
                for (tiler_name, tiler) in tilers.items()
                if tiler.supported(tiff_file)
            )
        except (TiffFileError, StopIteration):
            return None, None

    @classmethod
    def detect_format(
        cls,
        filepath: Union[str, Path, UPath],
        file_options: Optional[Dict[str, str]] = None,
    ) -> Optional[str]:
        format, supported_tiler = cls.get_tiler(filepath, file_options)
        return format



#    Copyright 2022-2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image implementation for 3Dhistech tiff files."""

from typing import List, Optional

from tifffile.tifffile import COMPRESSION, PHOTOMETRIC, TiffPage

from opentile.geometry import Size, SizeMm
from opentile.tiff_image import LockableFileHandle, NativeTiledTiffImage


class HistechTiffImage(NativeTiledTiffImage):
    def __init__(self, page: TiffPage, fh: LockableFileHandle, base_size: Size):
        """OpenTiledPage for 3DHistech Tiff image.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: LockableFileHandle
            Filehandler to read data from.
        base_size: Size
            Size of base level in pyramid.
        """
        super().__init__(page, fh)
        self._base_size = base_size
        self._pyramid_index = self._calculate_pyramidal_index(self._base_size)
        self._mpp = self._get_mpp_from_page()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._page}, {self._fh}, " f"{self._base_size})"

    @property
    def pixel_spacing(self) -> SizeMm:
        """Return pixel spacing in mm per pixel."""
        return self.mpp * 1000

    @property
    def supported_compressions(self) -> Optional[List[COMPRESSION]]:
        return None

    @property
    def mpp(self) -> SizeMm:
        """Return pixel spacing in um per pixel."""
        return self._mpp

    @property
    def photometric_interpretation(self) -> PHOTOMETRIC:
        photometric_interpretation = PHOTOMETRIC(self._page.photometric)
        if photometric_interpretation == PHOTOMETRIC.PALETTE:
            return PHOTOMETRIC.MINISBLACK
        return photometric_interpretation

    def _get_mpp_from_page(self) -> SizeMm:
        items_split = self._page.description.split("|")
        header = items_split.pop(0)
        items = {
            key: value
            for (key, value) in (item.split(" = ", 1) for item in items_split)
        }
        return (
            SizeMm(float(items["3dh_PixelSizeX"]), float(items["3dh_PixelSizeY"]))
            / 1000
            / 1000
        )



#    Copyright 2021-2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image implementations for svs tiff files."""

from typing import Dict, Iterator, List, Optional, Sequence, Tuple

import numpy as np
from imagecodecs import JPEG2K, JPEG8, jpeg2k_encode, jpeg8_encode, jpeg8_decode
from PIL import Image
from tifffile.tifffile import COMPRESSION, PHOTOMETRIC, TiffPage

from opentile.geometry import Point, Region, Size, SizeMm
from opentile.jpeg import Jpeg
from opentile.tiff_image import LockableFileHandle, NativeTiledTiffImage, TiffImage


class SvsStripedImage(TiffImage):
    _pyramid_index = 0

    def __init__(self, page: TiffPage, fh: LockableFileHandle, jpeg: Jpeg):
        """OpenTiledPage for jpeg striped Svs image, e.g. overview image.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: LockableFileHandle
            Filehandler to read data from.
        jpeg: Jpeg
            Jpeg instance to use.

        """
        super().__init__(page, fh, True)
        self._jpeg = jpeg

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._page}, {self._fh}, {self._jpeg}"

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        return None

    @property
    def supported_compressions(self) -> Optional[List[COMPRESSION]]:
        return [COMPRESSION.JPEG]

    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Return tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        if tile_position != (0, 0):
            raise ValueError("Non-tiled image, expected tile_position (0, 0)")
        indices = range(len(self.page.dataoffsets))
        scans = (self._read_frame(index) for index in indices)
        jpeg_tables = self.page.jpegtables
        frame = self._jpeg.concatenate_scans(
            scans, jpeg_tables, self._add_rgb_colorspace_fix
        )
        return frame

    def get_decoded_tile(self, tile_position: Tuple[int, int]) -> np.ndarray:
        """Return decoded tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        return jpeg8_decode(self.get_tile(tile_position))


class SvsLZWImage(TiffImage):
    _pyramid_index = 0

    def __init__(self, page: TiffPage, fh: LockableFileHandle, jpeg: Jpeg):
        """OpenTiledPage for lzw striped Svs image, e.g. label image.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: LockableFileHandle
            Filehandler to read data from.
        jpeg: Jpeg
            Jpeg instance to use.

        """
        super().__init__(page, fh)
        self._jpeg = jpeg

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._page}, {self._fh}, {self._jpeg}"

    @property
    def compression(self) -> COMPRESSION:
        """Return compression of page."""
        return COMPRESSION.JPEG

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        return None

    @property
    def supported_compressions(self) -> Optional[List[COMPRESSION]]:
        return [COMPRESSION.LZW]

    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Return tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        return jpeg8_encode(self.get_decoded_tile(tile_position))

    def get_decoded_tile(self, tile_position: Tuple[int, int]) -> np.ndarray:
        """Return decoded tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        if tile_position != (0, 0):
            raise ValueError("Non-tiled image, expected tile_position (0, 0)")

        tile = np.concatenate(
            [self._get_row(index) for index in range(len(self.page.dataoffsets))],
            axis=1,
        )
        return np.squeeze(tile)

    def _get_row(self, index: int) -> np.ndarray:
        row = self.page.decode(self._read_frame(index), index)[0]
        assert isinstance(row, np.ndarray)
        return row


class SvsTiledImage(NativeTiledTiffImage):
    def __init__(
        self,
        page: TiffPage,
        fh: LockableFileHandle,
        base_size: Size,
        base_mpp: SizeMm,
        parent: Optional[TiffImage] = None,
    ):
        """Svs Tiff tiled image.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: LockableFileHandle
            Filehandler to read data from.
        base_size: Size
            Size of base level in pyramid.
        base_mpp: SizeMm
            Mpp (um/pixel) for base level in pyramid.
        parent: Optional[TiffImage] = None
            Parent TiffImage
        """

        super().__init__(page, fh, True)
        self._base_size = base_size
        self._base_mpp = base_mpp
        self._pyramid_index = self._calculate_pyramidal_index(self._base_size)
        self._mpp = self._calculate_mpp(self._base_mpp)
        self._parent = parent
        (
            self._right_edge_corrupt,
            self._bottom_edge_corrupt,
        ) = self._detect_corrupt_edges()
        self._fixed_tiles: Dict[Point, bytes] = {}

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, "
            f"{self._base_size}, {self._base_mpp})"
        )

    @property
    def pixel_spacing(self) -> SizeMm:
        """Return pixel spacing in mm per pixel."""
        return self.mpp / 1000

    @property
    def supported_compressions(self) -> Optional[List[COMPRESSION]]:
        return [COMPRESSION.JPEG, COMPRESSION.APERIO_JP2000_RGB]

    @property
    def mpp(self) -> SizeMm:
        """Return pixel spacing in um per pixel."""
        return self._mpp

    @property
    def right_edge_corrupt(self) -> bool:
        return self._right_edge_corrupt

    @property
    def bottom_edge_corrupt(self) -> bool:
        return self._bottom_edge_corrupt

    def _detect_corrupt_edge(self, edge: Region) -> bool:
        """Return true if tiles at edge are corrupt (any tile has a frame
        that has 0 data length)
        Parameters
        ----------
        edge: Region
            Edge to check for corrupt tiles.

        Returns
        ----------
        bool:
            True if edge contains corrupt tiles.
        """
        for tile in edge.iterate_all():
            frame_index = self._tile_point_to_frame_index(tile)
            if self._page.databytecounts[frame_index] == 0:
                return True
        return False

    def _detect_corrupt_edges(self) -> Tuple[bool, bool]:
        """Returns tuple bool indiciting if right and/or bottom edge of page is
        corrupt. Bottom and right edge tiles in svs pyramid levels > 1 can
        sometimes corrupt. This manifests as tiles with 0-length (easy to
        detect) and tiles with scrambled image data (not so easy detect).

        Returns
        ----------
        Tuple[bool, bool]:
            Tuple indicating if right and/or bottom edge is corrupt.

        """
        if self.pyramid_index == 0:
            return False, False
        right_edge = Region(
            Point(self.tiled_size.width - 1, 0), Size(1, self.tiled_size.height - 1)
        )
        right_edge_corrupt = self._detect_corrupt_edge(right_edge)

        bottom_edge = Region(
            Point(0, self.tiled_size.height - 1), Size(self.tiled_size.width - 1, 1)
        )
        bottom_edge_corrupt = self._detect_corrupt_edge(bottom_edge)

        return right_edge_corrupt, bottom_edge_corrupt

    def _tile_is_at_right_edge(self, tile_point: Point) -> bool:
        """Return true if tile is at right edge of tiled image."""
        return tile_point.x == self.tiled_size.width - 1

    def _tile_is_at_bottom_edge(self, tile_point: Point) -> bool:
        """Return true if tile is at bottom edge of tiled image."""
        return tile_point.y == self.tiled_size.height - 1

    def _get_scaled_tile(self, tile_point: Point) -> bytes:
        """Create a tile by downscaling from a lower (higher resolution)
        level.

        Parameters
        ----------
        tile_point: Point
            Position of tile in this pyramid level.

        Returns
        ----------
        bytes:
            Scaled tile bytes.

        """
        if self._parent is None:
            raise ValueError("No parent level to get tiles from")
        scale = int(pow(2, self.pyramid_index - self._parent.pyramid_index))
        scaled_tile_region = Region(tile_point, Size(1, 1)) * scale

        # Get decoded tiles
        decoded_tiles = self._parent.get_decoded_tiles(
            [tile.to_tuple() for tile in scaled_tile_region.iterate_all()]
        )
        image_data = np.zeros(
            (self.tile_size * scale).to_tuple() + (3,), dtype=np.uint8
        )
        # Insert decoded_tiles into image_data
        for y in range(scale):
            for x in range(scale):
                image_data[
                    y * self.tile_size.width : (y + 1) * self.tile_size.width,
                    x * self.tile_size.height : (x + 1) * self.tile_size.height,
                ] = next(decoded_tiles)

        # Resize image_data using Pillow
        image: Image.Image = Image.fromarray(image_data)
        image = image.resize(
            self.tile_size.to_tuple(), resample=Image.Resampling.BILINEAR
        )

        # Return compressed image
        if self.compression == COMPRESSION.JPEG:
            if self._page.photometric == PHOTOMETRIC.RGB:
                colorspace = JPEG8.CS.RGB
            elif self._page.photometric == PHOTOMETRIC.YCBCR:
                colorspace = JPEG8.CS.YCbCr
            else:
                raise NotImplementedError("Non-supported color space")
            subsampling = self._page.subsampling
            return jpeg8_encode(
                np.array(image),
                level=95,
                colorspace=colorspace,
                subsampling=subsampling,
                lossless=False,
                bitspersample=8,
            )
        if self.compression == COMPRESSION.APERIO_JP2000_RGB:
            return jpeg2k_encode(
                np.array(image),
                level=80,
                codecformat=JPEG2K.CODEC.J2K,
                colorspace=JPEG2K.CLRSPC.SRGB,
                bitspersample=8,
                reversible=False,
                mct=True,
            )
        raise NotImplementedError("Non-supported compression")

    def _get_fixed_tile(self, tile_point: Point) -> bytes:
        """Get or create a fixed tile inplace for a corrupt tile.

        Parameters
        ----------
        tile_point: Point
            Position of tile to get.

        Returns
        ----------
        bytes:
            Fixed tile bytes.

        """
        if tile_point not in self._fixed_tiles:
            fixed_tile = self._get_scaled_tile(tile_point)
            self._fixed_tiles[tile_point] = fixed_tile
        return self._fixed_tiles[tile_point]

    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Return image bytes for tile at tile position. If tile is marked as
        corrupt, return a fixed tile. Add color space fix if jpeg compression.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        tile_point = Point.from_tuple(tile_position)
        if self._tile_is_corrupt(tile_point):
            return self._get_fixed_tile(tile_point)

        return super().get_tile(tile_position)

    def get_tiles(self, tile_positions: Sequence[Tuple[int, int]]) -> Iterator[bytes]:
        """Return list of image bytes for tiles at tile positions.

        Parameters
        ----------
        tile_positions: Sequence[Tuple[int, int]]
            Tile positions to get.

        Returns
        ----------
        Iterator[bytes]
            List of tile bytes.
        """
        tile_points = [
            Point.from_tuple(tile_position) for tile_position in tile_positions
        ]
        if any(self._tile_is_corrupt(tile_point) for tile_point in tile_points):
            return (self.get_tile(tile) for tile in tile_positions)
        return super().get_tiles(tile_positions)

    def _tile_is_corrupt(self, tile_point: Point) -> bool:
        """Return true if tile is corrupt

        Parameters
        ----------
        tile_point: Point
            Tile to check

        Returns
        ----------
        bool
            True if tile is corrupt.
        """
        return (
            self.right_edge_corrupt and self._tile_is_at_right_edge(tile_point)
        ) or (self.bottom_edge_corrupt and self._tile_is_at_bottom_edge(tile_point))



#    Copyright 2021-2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Metadata parser for ndpi files."""

from datetime import datetime
from functools import cached_property
from typing import Any, Dict, List, Optional

from tifffile.tifffile import TiffPage

from opentile.metadata import Metadata


def get_value_from_ndpi_comments(
    comments: str, value_name: str, value_type: Any
) -> Any:
    """Read value from ndpi comment string."""
    for line in comments.split("\n"):
        if value_name in line:
            value_string = line.split("=")[1]
            return value_type(value_string)


class NdpiMetadata(Metadata):
    def __init__(self, page: TiffPage):
        self._tags = page.tags
        if page.ndpi_tags is not None:
            self._ndpi_tags = page.ndpi_tags
        else:
            self._ndpi_tags = {}

    @cached_property
    def magnification(self) -> Optional[float]:
        try:
            return float(self._ndpi_tags["Magnification"])
        except (AttributeError, ValueError):
            return None

    @property
    def scanner_manufacturer(self) -> Optional[str]:
        return self._ndpi_tags.get("Make")

    @property
    def scanner_model(self) -> Optional[str]:
        return self._ndpi_tags.get("Model")

    @property
    def scanner_software_versions(self) -> Optional[List[str]]:
        software_version = self._ndpi_tags.get("Software")
        if software_version is not None:
            return [software_version]
        return None

    @property
    def scanner_serial_number(self) -> Optional[str]:
        return self._ndpi_tags.get("ScannerSerialNumber")

    @cached_property
    def aquisition_datetime(self) -> Optional[datetime]:
        datetime_tag = self._tags.get("DateTime")
        if datetime_tag is None:
            return None
        try:
            return datetime.strptime(datetime_tag.value, "%Y:%m:%d %H:%M:%S")
        except ValueError:
            return None

    @cached_property
    def properties(self) -> Dict[str, Any]:
        x_offset_from_slide_center = self._ndpi_tags.get("XOffsetFromSlideCenter")
        y_offset_from_slide_center = self._ndpi_tags.get("YOffsetFromSlideCenter")
        z_offset_from_slide_center = self._ndpi_tags.get("ZXOffsetFromSlideCenter")
        return {
            "x_offset_from_slide_center": x_offset_from_slide_center,
            "y_offset_from_slide_center": y_offset_from_slide_center,
            "z_offset_from_slide_center": z_offset_from_slide_center,
        }



#    Copyright 2021-2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Lossless jpeg handling."""

from pathlib import Path
from struct import pack, unpack
from typing import Iterator, List, Optional, Sequence, Tuple, Union

from turbojpeg import tjMCUHeight, tjMCUWidth

from opentile.geometry import Size
from opentile.jpeg.turbojpeg_patch import TurboJPEG_patch as TurboJPEG
from opentile.jpeg.turbojpeg_patch import find_turbojpeg_path


class JpegTagNotFound(Exception):
    """Raised when expected Jpeg tag was not found."""

    pass


class JpegCropError(Exception):
    """Raised when crop operation fails."""

    pass


class Jpeg:
    TAGS = {
        "tag marker": 0xFF,
        "start of image": 0xD8,
        "application default header": 0xE0,
        "quantization table": 0xDB,
        "start of frame": 0xC0,
        "huffman table": 0xC4,
        "start of scan": 0xDA,
        "end of image": 0xD9,
        "restart interval": 0xDD,
        "restart mark": 0xD0,
    }

    def __init__(self, turbo_path: Optional[Union[str, Path]] = None) -> None:
        if turbo_path is None:
            turbo_path = find_turbojpeg_path()
        self._turbo_jpeg = TurboJPEG(turbo_path)

    def get_mcu(self, frame: bytes) -> Size:
        """Return MCU size read from frame header.

        Parameters
        ----------
        frame: bytes
            Frame with header.

        Returns
        ----------
        bytes:
            MCU size in header.
        """
        _, _, subsampling, _ = self._turbo_jpeg.decode_header(frame)
        try:
            return Size(tjMCUWidth[subsampling], tjMCUHeight[subsampling])
        except IndexError:
            raise ValueError(f"Unknown subsampling {subsampling}.")

    def concatenate_fragments(self, fragments: Iterator[bytes], header: bytes) -> bytes:
        """Return frame created by vertically concatenating fragments.

        Parameters
        ----------
        fragments: Iterator[bytes]
            Iterator providing fragments to concatenate.
        header: bytes
            Heaeder for the frame.

        Returns
        ----------
        bytes:
            Concatenated frame in bytes.
        """
        frame = header
        for fragment_index, fragment in enumerate(fragments):
            if not (fragment[-2] == Jpeg.TAGS["tag marker"] and fragment[-1] != b"0"):
                raise JpegTagNotFound(
                    "Tag for end of scan or restart marker not found in scan"
                )
            frame += fragment[:-1]  # Do not include restart mark index
            frame += self.restart_mark(fragment_index)
        frame += self.end_of_image()
        return frame

    def concatenate_scans(
        self,
        scans: Iterator[bytes],
        jpeg_tables: Optional[bytes],
        rgb_colorspace_fix: bool = False,
    ) -> bytes:
        """Return frame created by horisontal concatenating scans. Scans must
        have the same header content, and only the last scan is allowed to have
        a different height.

        Parameters
        ----------
        scans: Iterator[bytes]
            Iterator providing scans to concatenate.
        jpeg_tables: Optional[bytes]
            Optional jpeg tables to insert into frame.
        rgb_colorspace_fix: bool = False
            If to apply rgb color space fix (for svs files).

        Returns
        ----------
        bytes:
            Concatenated frame in bytes.
        """
        frame = bytearray()
        image_size: Optional[Size] = None
        scan_size: Optional[Size] = None
        subsample: Optional[int] = None
        for scan_index, scan in enumerate(scans):
            width, height, _subsample, _ = self._turbo_jpeg.decode_header(scan)
            if image_size is None:
                image_size = Size(width, height)
                scan_size = Size(width, height)
                scan_start = 0
                subsample = _subsample
            else:
                image_size.height += height
                start_of_scan, length = self._find_tag(scan, self.start_of_scan())
                if start_of_scan is None or length is None:
                    raise JpegTagNotFound("Start of scan not found in header")
                scan_start = start_of_scan + length + 2

            frame += scan[scan_start:-2]
            frame += b"\xFF" + self.restart_mark(scan_index)

        frame[-2:] = self.end_of_image()

        if jpeg_tables is not None:
            if rgb_colorspace_fix:
                frame = self._add_jpeg_tables_and_rgb_color_space_fix(
                    frame, jpeg_tables
                )
            else:
                frame = self._add_jpeg_tables(frame, jpeg_tables)

        assert (
            image_size is not None and scan_size is not None and subsample is not None
        )
        frame = self._manipulate_header(
            frame, image_size, scan_size.area // self.subsample_to_mcu_size(subsample)
        )
        return bytes(frame)

    def fill_frame(self, frame: bytes, luminance: float) -> bytes:
        """Return frame filled with color from luminance.

        Parameters
        ----------
        frame: bytes
            Frame to fill.
        luminance: float
            Luminance to fill (0: black - 1: white).

        Returns
        ----------
        bytes:
            Frame with constant color from luminance.
        """
        return self._turbo_jpeg.fill_image(frame, luminance)

    def crop_multiple(
        self, frame: bytes, crop_parameters: Sequence[Tuple[int, int, int, int]]
    ) -> List[bytes]:
        """Crop multiple frames out of frame.

        Parameters
        ----------
        frame: bytes
            Frame to crop from.
        crop_parameters: Sequence[Tuple[int, int, int, int]]
            Parameters for each crop, specified as left position, top position,
            width, height.

        Returns
        ----------
        List[bytes]:
            Croped frames.
        """
        try:
            return self._turbo_jpeg.crop_multiple(frame, crop_parameters)
        except OSError as exception:
            raise JpegCropError(
                f"Crop of frame failed " f"with parameters {crop_parameters}"
            ) from exception

    @classmethod
    def add_jpeg_tables(
        cls,
        frame: bytes,
        jpeg_tables: bytes,
        apply_rgb_colorspace_fix: Optional[bool] = False,
    ) -> bytes:
        """Add jpeg tables to frame. Tables are inserted before 'start of
        scan'-tag, and leading 'start of image' and ending 'end of image' tags
        are removed from the header prior to insertion.

        Parameters
        ----------
        frame: bytes
            'Abbreviated' jpeg frame lacking jpeg tables.
        jpeg_tables: bytes
            Jpeg tables to add
        apply_rgb_colorspace_fix: Optional[bool] = False
            If to also apply

        Returns
        ----------
        bytes:
            'Interchange' jpeg frame containing jpeg tables.

        """

        if apply_rgb_colorspace_fix:
            frame = cls._add_jpeg_tables_and_rgb_color_space_fix(
                bytearray(frame), jpeg_tables
            )
        else:
            frame = cls._add_jpeg_tables(bytearray(frame), jpeg_tables)
        return bytes(frame)

    @classmethod
    def manipulate_header(
        cls,
        frame: bytes,
        image_size: Optional[Size] = None,
        restart_interval: Optional[int] = None,
    ) -> bytes:
        """Return frame with changed header to reflect changed image size
        or restart interval.

        Parameters
        ----------
        frame: bytes
            Frame with header to update.
        image_size: Optional[Size] = None
            Image size to update header with.
        restart_interval: Optional[int] = None
            Restart interval to update header with.

        Returns
        ----------
        bytes:
            Frame with updated header.

        """
        return bytes(
            cls._manipulate_header(bytearray(frame), image_size, restart_interval)
        )

    @classmethod
    def start_of_frame(cls) -> bytes:
        """Return bytes representing a start of frame tag."""
        return bytes([cls.TAGS["tag marker"], cls.TAGS["start of frame"]])

    @classmethod
    def start_of_scan(cls) -> bytes:
        """Return bytes representing a start of scan tag."""
        return bytes([cls.TAGS["tag marker"], cls.TAGS["start of scan"]])

    @classmethod
    def end_of_image(cls) -> bytes:
        """Return bytes representing a end of image tag."""
        return bytes([cls.TAGS["tag marker"], cls.TAGS["end of image"]])

    @classmethod
    def restart_mark(cls, index: int) -> bytes:
        """Return bytes representing a restart marker of index (0-7), without
        the prefixing tag (0xFF)."""
        return bytes([cls.TAGS["restart mark"] + index % 8])

    @classmethod
    def restart_interval(cls) -> bytes:
        return bytes([cls.TAGS["tag marker"], cls.TAGS["restart interval"]])

    @staticmethod
    def code_short(value: int) -> bytes:
        return pack(">H", value)

    @staticmethod
    def subsample_to_mcu_size(subsample: int) -> int:
        return tjMCUWidth[subsample] * tjMCUHeight[subsample]

    @staticmethod
    def _find_tag(
        frame: Union[bytes, bytearray], tag: bytes
    ) -> Tuple[Optional[int], Optional[int]]:
        """Return first index and length of payload of tag in header.

        Parameters
        ----------
        frame: bytes
            Frame with header to search.
        tag: bytes
            Tag to search for.

        Returns
        ----------
        Tuple[Optional[int], Optional[int]]:
            Position of tag in header and length of payload.
        """
        index = frame.find(tag)
        if index != -1:
            (length,) = unpack(">H", frame[index + 2 : index + 4])
            return index, length

        return None, None

    @classmethod
    def _manipulate_header(
        cls,
        frame: bytearray,
        size: Optional[Size] = None,
        restart_interval: Optional[int] = None,
    ) -> bytearray:
        """Return manipulated header with changed pixel size (width, height)
        and/or restart interval.

        Parameters
        ----------
        frame: bytearray
            Frame to manipulate header of.
        size: Optional[Size] = None
            Pixel size to insert into header.
        restart_interval: Optional[int] = None
            Restart interval to code into header.

        Returns
        ----------
        bytearray:
            Manipulated header.
        """
        if size is not None:
            start_of_frame_index, _ = cls._find_tag(frame, cls.start_of_frame())
            if start_of_frame_index is None:
                raise JpegTagNotFound("Start of frame tag not found in header")
            size_index = start_of_frame_index + 5
            frame[size_index : size_index + 2] = cls.code_short(size.height)
            frame[size_index + 2 : size_index + 4] = cls.code_short(size.width)

        if restart_interval is not None:
            restart_payload = cls.code_short(restart_interval)
            restart_index, _ = cls._find_tag(frame, cls.restart_interval())
            if restart_index is not None:
                # Modify existing restart tag
                payload_index = restart_index + 4
                frame[payload_index : payload_index + 2] = restart_payload
            else:
                # Make and insert new restart tag
                start_of_scan_index, _ = cls._find_tag(frame, cls.start_of_scan())
                if start_of_scan_index is None:
                    raise JpegTagNotFound("Start of scan tag not found in header")
                frame[start_of_scan_index:start_of_scan_index] = (
                    cls.restart_interval() + cls.code_short(4) + restart_payload
                )
        return frame

    @classmethod
    def _add_jpeg_tables(
        cls,
        frame: bytearray,
        jpegtables: bytes,
    ) -> bytearray:
        """Add jpeg tables to frame."""
        start_of_scan = frame.find(cls.start_of_scan())
        frame[start_of_scan:start_of_scan] = jpegtables[2:-2]
        return frame

    @classmethod
    def _add_jpeg_tables_and_rgb_color_space_fix(
        cls,
        frame: bytearray,
        jpegtables: bytes,
    ) -> bytearray:
        """Add jpeg tables to frame and add Adobe APP14 marker with transform
        flag 0 indicating image is encoded as RGB (not YCbCr)."""
        start_of_scan = frame.find(cls.start_of_scan())
        frame[start_of_scan:start_of_scan] = (
            jpegtables[2:-2]
            + b"\xFF\xEE\x00\x0E\x41\x64\x6F\x62"
            + b"\x65\x00\x64\x80\x00\x00\x00\x00"
        )
        return frame



#    Copyright 2021-2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Tiler for reading tiles from svs files."""

from functools import lru_cache
from pathlib import Path
from typing import Optional, Union

from tifffile.tifffile import TiffFile, TiffPageSeries
from upath import UPath

from opentile.formats.svs.svs_image import SvsLZWImage, SvsStripedImage, SvsTiledImage
from opentile.formats.svs.svs_metadata import SvsMetadata
from opentile.geometry import SizeMm
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata
from opentile.tiler import TiffImage, Tiler


class SvsTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path, UPath],
        turbo_path: Optional[Union[str, Path]] = None,
    ):
        """Tiler for svs file.

        Parameters
        ----------
        filepath: Union[str, Path, UPath]
            Filepath to a svs TiffFile.
        """
        super().__init__(filepath)
        self._jpeg = Jpeg(turbo_path)
        if "InterColorProfile" in self._tiff_file.pages.first.tags:
            icc_profile = self._tiff_file.pages.first.tags["InterColorProfile"].value
            assert isinstance(icc_profile, bytes) or icc_profile is None
            self._icc_profile = icc_profile
        self._metadata = SvsMetadata(self.base_page)
        self._base_mpp = SizeMm(self._metadata.mpp, self._metadata.mpp)

    @property
    def base_mpp(self) -> SizeMm:
        """Return pixel spacing in um/pixel for base level."""
        return self._base_mpp

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return tiff_file.is_svs

    @staticmethod
    def _is_level_series(series: TiffPageSeries) -> bool:
        return series.name == "Baseline"

    @staticmethod
    def _is_label_series(series: TiffPageSeries) -> bool:
        return series.name == "Label"

    @staticmethod
    def _is_overview_series(series: TiffPageSeries) -> bool:
        return series.name == "Macro"

    @lru_cache(None)
    def get_level(self, level: int, page: int = 0) -> TiffImage:
        series = self._level_series_index
        if level > 0:
            parent = self.get_level(level - 1, page)
        else:
            parent = None
        svs_page = SvsTiledImage(
            self._get_tiff_page(series, level, page),
            self._fh,
            self.base_size,
            self.base_mpp,
            parent,
        )
        return svs_page

    @lru_cache(None)
    def get_label(self, page: int = 0) -> TiffImage:
        if self._label_series_index is None:
            raise ValueError("No label detected in file")
        return SvsLZWImage(
            self._get_tiff_page(self._label_series_index, 0, page), self._fh, self._jpeg
        )

    @lru_cache(None)
    def get_overview(self, page: int = 0) -> TiffImage:
        if self._overview_series_index is None:
            raise ValueError("No overview detected in file")
        return SvsStripedImage(
            self._get_tiff_page(self._overview_series_index, 0, page),
            self._fh,
            self._jpeg,
        )
